"""
In this package, we provide implementations of metaheuristic algorithms.

The black-box methods are given directly in this package, more specialized
algorithms will be placed in sub-packages corresponding to their requirements
for the search- and solution space.
"""
